#include <iostream>
#include <stdlib.h>
#include <unistd.h>

#include "sqldb.h"
#include "strmanip.h"

using std::endl;
using std::cerr;

// Fuzzy compare function using levenshtein string distance
static void fuzzy_like(sqlite_func *context, int arg, const char **argv)
{
  if (argv[0] == 0 || argv[1] ==0)
      return;
  int len1 = strlen(argv[0]);
  int len2 = strlen(argv[1]);
  int distance = lev_edit_distance(len1, argv[0], len2, argv[1]);
  sqlite_set_result_int(context, (len1 + len2) / 8 > distance);
}

SqlDb::SqlDb() : nrow(0), ncol(0), resultp(0), errmsg(0)
{
    string dbpath = getenv("HOME");    
    db = sqlite_open(dbpath.append("/.imms/imms.db").c_str(), 600, &errmsg);
    if (!db)
        cerr << "Failed to open database '" << dbpath << "'" << endl;
    else
    {
        // Override the default LIKE operation with fuzzy_like
        sqlite_create_function(db, "like", 2, fuzzy_like, 0);
    }

    sql_create_table();
}

SqlDb::~SqlDb()
{
    sqlite_free_table(resultp);
    if (db)
        sqlite_close(db);
}

void SqlDb::sql_create_table()
{
    run_query(
            "CREATE TABLE 'Library' ("
                "'sid' INTEGER NOT NULL, "
                "'path' VARCHAR(4096) UNIQUE NOT NULL, "
                "'modtime' TIMESTAMP NOT NULL, "
                "'checksum' VARCHAR(34) NOT NULL);");

    run_query(
            "CREATE TABLE 'Rating' ("
                "'sid' INTEGER UNIQUE NOT NULL, "
                "'rating' INTEGER NOT NULL);");

    run_query(
            "CREATE TABLE 'Info' ("
                "'sid' INTEGER UNIQUE NOT NULL," 
                "'artist' VARCHAR(255) NOT NULL, "
                "'title' VARCHAR(255) NOT NULL);");

    run_query(
            "CREATE TABLE 'Last' ("
                "'artist' VARCHAR(255) NOT NULL, "
                "'title' VARCHAR(255) NOT NULL, "
                "'last' TIMESTAMP);");

    run_query(
            "CREATE TABLE 'UnknownLast' ("
                "'sid' INTEGER UNIQUE NOT NULL, "
                "'last' TIMESTAMP NOT NULL);");
}

int SqlDb::identify(const string &path, time_t modtime)
{
    title = artist = "";

    select_query(
            "SELECT sid, modtime FROM 'Library' "
            "WHERE path = '" + escape_string(path) + "';");

    sid = -1;

    if (nrow > 0)
    {
        time_t last_modtime = atol(resultp[ncol + 1]);
        sid = atoi(resultp[ncol]);

        if (modtime == last_modtime)
        {
#ifdef DEBUG
            cerr << " >> identify returning " << sid << endl;
#endif
            return sid;
        }
    }

    return -1;
}

int SqlDb::identify(const string &path, time_t modtime, const string &checksum)
{
    // old path but modtime has changed - update checksum
    if (nrow)
    {
        run_query(
                "UPDATE 'Library' SET "
                    "modtime = '" + itos(modtime) + "', "
                    "checksum = '" + checksum + "' "
                "WHERE path = '" + escape_string(path) + "';");
        return sid;
    }

    // moved or new file and path needs updating

    sid = -1;

    select_query(
            "SELECT sid, path FROM 'Library' "
            "WHERE checksum = '" + checksum + "';");

    if (nrow)
    {
        run_query(
                "DELETE FROM 'Info' "
                "WHERE sid = '" + string(resultp[ncol]) + "';");

        // Check if any of the old paths no longer exist
        // so that we can reuse its slot
        
        int col = 0;

        while (resultp[col += ncol])
        {
            sid = atoi(resultp[ncol]);

            if (access(resultp[col + 1], F_OK))
            {
                run_query(
                        "UPDATE 'Library' SET "
                            "path = '" + escape_string(path) + "', "
                            "modtime = '" + itos(modtime) + "' "
                        "WHERE path = '" + resultp[col + 1] + "';");

                return sid;
            }
        }
    }
    else
    {
        // figure out what the next sid should be
        select_query("SELECT max(sid) FROM Library;");
        sid = resultp[1] ? atoi(resultp[1]) + 1 : 1;
    }

    // new file - insert into the database
    run_query(
            "INSERT INTO 'Library' ('sid', 'path', 'modtime', 'checksum') "
            "VALUES ('" +
                itos(sid) + "', '" +
                escape_string(path) + "', '" +
                itos(modtime) + "', '" +
                checksum + "'"
            ");");

    run_query(
            "INSERT INTO 'Rating' ('sid', 'rating') "
            "VALUES ('" + itos(sid) + "', '-1');");
    run_query(
            "INSERT INTO 'UnknownLast' ('sid', 'last') "
            "VALUES ('" + itos(sid) + "', '0');");

    return sid;
}

int SqlDb::check_artist(const string &artist)
{
    select_query(
            "SELECT artist FROM 'Info' "
            "WHERE artist LIKE '" + artist + "';");
    return nrow;
}

int SqlDb::check_title(const string &title)
{
    select_query(
            "SELECT title FROM 'Info' "
            "WHERE artist = '" + artist + "' AND title LIKE '" + title + "';");
    return nrow;
}

string SqlDb::get_artist()
{
    if (sid == -1)
        return "";

    select_query(
            "SELECT artist FROM 'Info' "
            "WHERE sid = '" + itos(sid) + "';");

    return artist = nrow > 0 ? resultp[1] : "";
}

string SqlDb::get_title()
{
    if (sid == -1)
        return "";

    select_query(
            "SELECT title FROM 'Info' "
            "WHERE sid = '" + itos(sid) + "';");

    title = nrow > 0 ? resultp[1] : "";
#ifdef DEBUG
    cerr << " >> from cache: " << endl;
    cerr << " >>     title   = '" << title << "'" << endl;
    cerr << " >>     artist  = '" << artist << "'" << endl;
#endif
    return title;
}

time_t SqlDb::get_last()
{
    if (sid == -1)
        return 0;

    if (artist != "" && title != "")
    {
        select_query(
                "SELECT last FROM 'Last' "
                "WHERE artist = '" + artist + "' AND title = '" + title + "';");
    }
    else
    {
        select_query(
                "SELECT last FROM 'UnknownLast' "
                "WHERE sid = '" + itos(sid) + "';");
    }

    return resultp[1] ? atol(resultp[1]) : 0;
}

int SqlDb::get_rating()
{
    if (sid == -1)
        return -1;

    select_query(
            "SELECT rating FROM 'Rating' "
            "WHERE sid = '" + itos(sid) + "';");

    return nrow > 0 ? atoi(resultp[1]) : -1;
}

void SqlDb::set_artist(const string &_artist)
{
    if (sid == -1)
        return;

    artist = (nrow > 0 && resultp[1]) ? resultp[1] : _artist;
}

void SqlDb::set_title(const string &_title)
{
    if (sid == -1)
        return;

    title = (nrow > 0 && resultp[1]) ? resultp[1] : _title;

    run_query(
            "INSERT INTO 'Info' ('sid', 'artist', 'title') "
            "VALUES ('" + itos(sid) + "', '" + artist + "', '" + title + "');");

    select_query(
            "SELECT last FROM 'UnknownLast' "
            "WHERE sid = '" + itos(sid) + "';");

    run_query(
            "INSERT INTO 'Last' ('last', 'artist', 'title') "
            "VALUES ('" +
                string(resultp[1] ? resultp[1] : "0") + "', '" +
                artist + "', '" +
                title + "');");

    run_query(
            "DELETE FROM 'UnknownLast' "
            "WHERE sid = '" + itos(sid) + "';");

#ifdef DEBUG
    cerr << " >> cached as: " << endl;
    cerr << " >>     title   = '" << title << "'" << endl;
    cerr << " >>     artist  = '" << artist << "'" << endl;
#endif
}

void SqlDb::set_last(time_t last)
{
    if (sid == -1)
        return;

    // known
    if (artist != "" && title != "")
    {
        run_query(
                "UPDATE 'Last' SET last = '" + itos(last) + "' "
                "WHERE artist = '" + artist + "' AND title = '" + title + "';");  
    }
    // unknown
    else
    {
        run_query(
                "UPDATE 'UnknownLast' SET last = '" + itos(last) + "' "
                "WHERE sid = '" + itos(sid) + "';");
    }
}

void SqlDb::set_rating(int rating)
{
    if (sid == -1)
        return;

    run_query(
            "UPDATE 'Rating' SET rating = '" + itos(rating) + "' "
            "WHERE sid = '" + itos(sid) + "';");
}

void SqlDb::run_query(const string &query)
{
    if (db)
    {
        sqlite_exec(db, query.c_str(), 0, 0, &errmsg);
        handle_error(query);
    }
    else
    {
        cerr << "Database not open!" << endl;
    }
}

void SqlDb::select_query(const string &query)
{
    if (db)
    {
        sqlite_free_table(resultp);

        sqlite_get_table(db, query.c_str(), &resultp, &nrow, &ncol, &errmsg);
        handle_error(query);
    }
    else
    {
        cerr << "Database not open!" << endl;
    }
}

void SqlDb::handle_error(const string &query)
{
    if (errmsg && !strstr(errmsg, "exists") && !strstr(errmsg, "unique"))
    {
        cerr << errmsg << endl;
        cerr << "while executing: " << query << endl;
    }
    free(errmsg);
    errmsg = 0;
}
